import cv2
import numpy as np
from scipy.ndimage import label
import math
from skimage.morphology import skeletonize


# 常量
CANVAS_SHAPE = (384, 384)
DILATED_KERNEL = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
CLOSE_KERNEL = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (11, 11))
STRUCTURE_4 = np.array([[0, 1, 0],
                        [1, 1, 1],
                        [0, 1, 0]])

STRUCTURE_8 = np.array([[1, 1, 1],
                        [1, 1, 1],
                        [1, 1, 1]])


def get_canvas(channels=1):
    """ 获取一张空画布图像 """
    if channels > 1:
        return np.zeros([*CANVAS_SHAPE, channels], dtype='uint8')
    return np.zeros(CANVAS_SHAPE, dtype='uint8')


def dilated(image, kernel=None, iteration=1):
    """ 膨胀运算 """
    dilated_image = cv2.dilate(image, DILATED_KERNEL if kernel is None else kernel, iterations=iteration)
    return dilated_image


def close(image, kernel=None, iteration=1):
    """ 闭运算 """
    closed_image = cv2.morphologyEx(image, cv2.MORPH_CLOSE, kernel=CLOSE_KERNEL if kernel is None else kernel,
                                    iterations=iteration)
    return closed_image


def distance(p1, p2):
    """ 计算欧氏距离 """
    p1, p2 = np.array(p1), np.array(p2)
    return np.linalg.norm(p1 - p2)


def split(image, split_skeleton=False):
    """ 分割连通性区域 """
    image = image > 0
    if not split_skeleton:
        arrays, num = label(image, structure=STRUCTURE_4)
    else:
        arrays, num = label(image, structure=STRUCTURE_8)

    segments = []
    for i in range(1, num + 1):
        component_image = np.where(arrays == i, 1, 0)
        component_image = component_image * 255
        component_image = component_image.astype('uint8')
        segments.append(component_image)

    return segments, num


def calculate_angle(p1, p2, p3):
    """ 计算近似曲线的拐角 """
    v1 = p1 - p2
    v2 = p3 - p2
    try:
        angle = math.degrees(math.acos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))))
    except ValueError:
        angle = 180
    return angle


def skeletonize_(image, min_area=8):
    """ 将图片骨架化 """
    image = image > 0
    skeleton = skeletonize(image)
    skeleton = skeleton.astype('uint8')
    skeleton = skeleton * 255

    # 设置边缘像素为0
    skeleton[0, :] = 0  # 上边缘
    skeleton[-1, :] = 0  # 下边缘
    skeleton[:, 0] = 0  # 左边缘
    skeleton[:, -1] = 0  # 右边缘

    # 过滤噪点
    if min_area > 0:
        segments, _ = split(skeleton, True)

        for segment in segments:
            if cv2.countNonZero(segment) < min_area:
                skeleton[segment > 0] = 0

    return skeleton



import numpy as np
import cv2


from aiccm.utils.morphology import skeletonize_, get_canvas, dilated, split



END_KERNEL = np.array([[1, 1, 1],
                       [1, 10, 1],
                       [1, 1, 1]])



def get_points(skeleton):
    """ 查找骨架图中的分支点 """
    neighbors = cv2.filter2D((skeleton > 0).astype('uint8'), -1, END_KERNEL)
    nodes_mask = neighbors >= 13
    ends_mask = neighbors == 11

    skeleton2 = skeleton.copy()
    skeleton2[nodes_mask] = 0

    segments, _ = split(skeleton2, True)

    for segment in segments:
        if cv2.countNonZero(segment) < 4 and cv2.countNonZero(
                cv2.bitwise_and(segment, (ends_mask * 255).astype('uint8'))):
            skeleton[segment > 0] = 0

    skeleton = skeletonize_(skeleton, min_area=-1)

    neighbors = cv2.filter2D((skeleton > 0).astype('uint8'), -1, END_KERNEL)
    nodes_mask = neighbors >= 13
    ends_mask = neighbors == 11

    points = get_canvas()
    points[nodes_mask] = 255
    points = dilated(points, iteration=3)
    points = cv2.bitwise_and(points, skeleton)

    end_points = get_canvas()
    end_points[ends_mask] = 255

    return points, end_points




from aiccm.utils.io import load_ccm_image
from aiccm.segment.detector import segmenter
from aiccm.utils.morphology import skeletonize_, split

from .point import get_points
from .instance import get_instance
from .graph import get_trunk
from .concat import get_nerve
from .draw import draw_result_image


class Processor:
    def __init__(self):
        self.image = None
        self.images = {
            'binary': None,
            'result': None,
        }

        self.model_path = None

        self.segments = []
        self.nodes = []
        self.nerves = []



    def load_image(self, image):
        self.image = image.copy()
        self.segments = []
        self.nodes = []
        self.nerves = []

    def process(self, trunk_rec=True):
        binary = segmenter(self.image)
        self.images['binary'] = binary
        skeleton = skeletonize_(binary)
        blocks, _ = split(skeleton, split_skeleton=True)
        for block in blocks:
            points, end_points = get_points(block)
            instance = get_instance(binary, block, points, end_points)
            segments, nodes = instance
            self.segments.extend(segments)
            self.nodes.extend(nodes)

            if trunk_rec:
                trunks = get_trunk(*instance)
                if trunks:
                    for sis, nis in trunks:
                        for si in sis:
                            segments[si].class_segment = 'main'
                        self.nerves.append(get_nerve([segments[si] for si in sis], [nodes[ni] for ni in nis]))

        # 重新编号
        for i, segment in enumerate(self.segments):
            segment.index = i

        for i, node in enumerate(self.nodes):
            node.index = i

        self.images['result'] = draw_result_image(self.segments, self.nodes, self.image)







from aiccm.utils.morphology import skeletonize_, get_canvas, split, dilated

import cv2
import numpy as np


class NerveSegment:
    def __init__(self, index, body, bone, class_='Segment'):
        self.index = index
        self.body = body
        self.bone = bone
        self.class_ = class_  # 'Segment' or 'Node'
        self.neighbors = []
        self.class_node = None  # 'end' or 'branch'
        self.class_segment = 'branch'  # 'main' or 'branch'

        # property
        self._area = None
        self._area_weighted = None
        self._length = None
        self._width = None
        self._width_weighted = None
        self._ends = None
        self._center = None

    def __repr__(self):
        return f'{self.class_} {self.index}'

    @property
    def center(self):
        """ 轮廓中心点 """
        if self._center is None:
            indices = np.argwhere(self.bone != 0)
            center = np.mean(indices, axis=0)
            self._center = int(center[1]), int(center[0])
        return self._center  # As (x, y)

    @property
    def length(self):
        """ 长度，即骨架线长度 """
        # https://blog.csdn.net/oYeZhou/article/details/130575866
        if self._length is None:
            length = 0
            contours, _ = cv2.findContours(self.bone, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            # 计算长度
            for c in contours:
                length += cv2.arcLength(c, True) / 2
            self._length = length
        return self._length

    @property
    def area(self):
        """ 面积，即像素点个数之和 """
        if self._area is None:
            self._area = cv2.countNonZero(self.body)
        return self._area

    @property
    def width(self):
        """ 宽度，即面积除以长度 """
        if self._width is None:
            self._width = self.area / self.length
        return self._width


class NerveContainer:
    """ Nerve对象容器，便于索引查找 """
    def __init__(self, instance):
        self.instance = instance


    def __repr__(self):
        return str([i.index for i in self.instance])

    def __getitem__(self, item):
        try:
            return [x for x in self.instance if x.index == item][0]
        except IndexError:
            raise IndexError('找不到该索引对应的对象.')

    def __iter__(self):
        return iter(self.instance)


def _deconvolution(skeleton, image, threshold=3):
    """ 利用距离变换，根据骨架反卷积出原图像像素 """
    skeleton = skeleton > 0
    skeleton = skeleton.astype('bool')

    dist_transform = cv2.distanceTransform((1 - skeleton).astype(np.uint8), cv2.DIST_L2, 5)
    distance_threshold = threshold
    selected_pixels = (dist_transform <= distance_threshold) & image

    segments, num = split(selected_pixels, False)
    if num == 1:
        return selected_pixels
    else:
        areas = [cv2.countNonZero(s) for s in segments]
        return segments[np.argmax(areas)]


def _edge(image):
    """ 获取一个连通性区域的外层轮廓 """
    image2 = dilated(image, np.ones((3, 3), dtype='uint8'))
    edge = np.bitwise_xor(image, image2)
    return edge



def get_instance(image, skeleton, points, end_points):
    """ 生成神经段或节点实例 """
    segments_list, nodes_list = [], []

    canvas = get_canvas()
    canvas[skeleton > 0] = 255
    canvas[points > 0] = 0
    segments, _ = split(canvas, split_skeleton=True)
    for i, s in enumerate(segments):
        segment = _deconvolution(s, image)
        segments_list.append(NerveSegment(
            i,
            segment.copy(),
            s.copy(),
            'Segment')
        )

    points = cv2.bitwise_or(points, end_points)
    nodes, _ = split(points, split_skeleton=True)
    for i, n in enumerate(nodes):
        node = NerveSegment(
            i,
            None,
            n,
            'Node'
        )

        node.class_node = 'end' if cv2.countNonZero(cv2.bitwise_and(node.bone, end_points)) else 'branch'
        nodes_list.append(node)
        edge = _edge(n)
        # 邻居识别
        for segment in segments_list:
            intersection = cv2.bitwise_and(edge, segment.bone)
            if cv2.countNonZero(intersection) > 0:
                node.neighbors.append(segment)
                if node not in segment.neighbors:
                    segment.neighbors.append(node)


    return NerveContainer(segments_list), NerveContainer(nodes_list)



from aiccm.utils.morphology import get_canvas
import cv2
from typing import Literal

DIGITAL_TYPE = Literal['none', 'both', 'segments', 'node']
SHOW_TYPE = Literal['bone', 'body']


def _draw_digit(image, segments=None, nodes=None):
    """ 附加数字 """
    canvas = image.copy()

    if segments is not None:
        for segment in segments:
            cv2.putText(canvas, str(segment.index), segment.center, cv2.FONT_HERSHEY_SIMPLEX, .3, (0, 255, 0), 1,
                        cv2.LINE_AA)

    if nodes is not None:
        for node in nodes:
            cv2.putText(canvas, str(node.index), node.center, cv2.FONT_HERSHEY_SIMPLEX, .3, (255, 255, 0), 1,
                        cv2.LINE_AA)

    return canvas


def draw_digit_image(segments, nodes, image=None):
    """ 创建带数字的神经骨架图 """
    canvas = get_canvas(3) if image is None else image.copy()

    for segment in segments:
        canvas[segment.bone > 0] = (255, 255, 255)
        cv2.putText(canvas, str(segment.index), segment.center, cv2.FONT_HERSHEY_SIMPLEX, .3, (0, 255, 0), 1,
                    cv2.LINE_AA)

    for node in nodes:
        canvas[node.bone > 0] = (0, 0, 255)
        cv2.putText(canvas, str(node.index), node.center, cv2.FONT_HERSHEY_SIMPLEX, .3, (255, 255, 0), 1, cv2.LINE_AA)

    return canvas


def draw_result_image(segments, nodes, image=None, digital: DIGITAL_TYPE = 'none', show: SHOW_TYPE = 'bone'):
    """ 创建用于结果展示的图像，即以原图为背景，红色骨架代表主干，蓝色骨架代表分支，绿色圆点代表分支点，黄色圆点代表端点 """
    canvas = get_canvas(3) if image is None else image.copy()

    for segment in segments:
        if show == 'bone':
            canvas[segment.bone > 0] = (0, 0, 255) if segment.class_segment == 'main' else (255, 0, 0)
        elif show == 'body':
            canvas[segment.body > 0] = (0, 0, 255) if segment.class_segment == 'main' else (255, 0, 0)

    for node in nodes:
        if show == 'bone' and node.class_node == 'branch':
            cv2.circle(canvas, node.center, 2, (0, 255, 0), -1)
        elif show == 'body' and node.class_node == 'branch':
            cv2.circle(canvas, node.center, 4, (0, 255, 0), -1)


    if digital == 'both':
        params = (segments, nodes)
    elif digital == 'segments':
        params = (segments,)
    elif digital == 'nodes':
        params = (nodes,)
    else:
        params = ()

    canvas = _draw_digit(canvas, *params)

    return canvas


def draw_trunk_image(segments, image=None):
    """ 创建仅标注主干的图像 """
    canvas = get_canvas(3) if image is None else image.copy()
    for segment in [s for s in segments if s.class_segment == 'main']:
        canvas[segment.bone > 0] = (0, 0, 255)

    return canvas



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from collections.abc import Sequence

import warnings


class ComponentManager:
    """
    Implement a manager class to add the new component properly.
    The component can be added as either class or function type.

    Args:
        name (str): The name of component.

    Returns:
        A callable object of ComponentManager.

    Examples 1:

        from paddleseg.cvlibs.manager import ComponentManager

        model_manager = ComponentManager()

        class AlexNet: ...
        class ResNet: ...

        model_manager.add_component(AlexNet)
        model_manager.add_component(ResNet)

        # Or pass a sequence alliteratively:
        model_manager.add_component([AlexNet, ResNet])
        print(model_manager.components_dict)
        # {'AlexNet': <class '__main__.AlexNet'>, 'ResNet': <class '__main__.ResNet'>}

    Examples 2:

        # Or an easier way, using it as a Python decorator, while just add it above the class declaration.
        from paddleseg.cvlibs.manager import ComponentManager

        model_manager = ComponentManager()

        @model_manager.add_component
        class AlexNet: ...

        @model_manager.add_component
        class ResNet: ...

        print(model_manager.components_dict)
        # {'AlexNet': <class '__main__.AlexNet'>, 'ResNet': <class '__main__.ResNet'>}
    """

    def __init__(self, name=None):
        self._components_dict = dict()
        self._name = name

    def __len__(self):
        return len(self._components_dict)

    def __repr__(self):
        name_str = self._name if self._name else self.__class__.__name__
        return "{}:{}".format(name_str, list(self._components_dict.keys()))

    def __getitem__(self, item):
        if item not in self._components_dict.keys():
            raise KeyError("{} does not exist in availabel {}".format(item,
                                                                      self))
        return self._components_dict[item]

    @property
    def components_dict(self):
        return self._components_dict

    @property
    def name(self):
        return self._name

    def _add_single_component(self, component):
        """
        Add a single component into the corresponding manager.

        Args:
            component (function|class): A new component.

        Raises:
            TypeError: When `component` is neither class nor function.
            KeyError: When `component` was added already.
        """

        # Currently only support class or function type
        if not (inspect.isclass(component) or inspect.isfunction(component)):
            raise TypeError("Expect class/function type, but received {}".
                            format(type(component)))

        # Obtain the internal name of the component
        component_name = component.__name__

        # Check whether the component was added already
        if component_name in self._components_dict.keys():
            warnings.warn("{} exists already! It is now updated to {} !!!".
                          format(component_name, component))
            self._components_dict[component_name] = component

        else:
            # Take the internal name of the component as its key
            self._components_dict[component_name] = component

    def add_component(self, components):
        """
        Add component(s) into the corresponding manager.

        Args:
            components (function|class|list|tuple): Support four types of components.

        Returns:
            components (function|class|list|tuple): Same with input components.
        """

        # Check whether the type is a sequence
        if isinstance(components, Sequence):
            for component in components:
                self._add_single_component(component)
        else:
            component = components
            self._add_single_component(component)

        return components


MODELS = ComponentManager("models")
BACKBONES = ComponentManager("backbones")
DATASETS = ComponentManager("datasets")
TRANSFORMS = ComponentManager("transforms")
LOSSES = ComponentManager("losses")
OPTIMIZERS = ComponentManager("optimizers")



class Nerve:
    def __init__(self, segments, nodes):
        self.segments = segments
        self.nodes = nodes

    def __repr__(self):
        return f"""Nerve
        Segments: {[s.index for s in self.segments]}
        Nodes: {[n.index for n in self.nodes]}
        """


def get_nerve(segments, nodes):
    """ 将NerveSegment拼接为Nerve """
    return Nerve(segments, nodes)







import itertools
from functools import lru_cache

from .instance import NerveContainer
from .point import get_points

from aiccm.utils.morphology import skeletonize_, distance, get_canvas, close, calculate_angle, split

import cv2
import numpy as np
import networkx as nx


class Trunk:
    def __init__(self, segments, nodes, graph, start_node, end_node):
        self.segments = segments
        self.nodes = nodes
        self.graph = graph
        self.start_node = start_node
        self.end_node = end_node

        self._body = None
        self._bone = None
        self._corner = None

    def __repr__(self):
        return f"""Trunk:
        Segments: {self.segments}
        Nodes: {self.nodes}
        Start: {self.start_node}, End: {self.end_node}
        Length: {self.length}, Linearity: {self.linearity}
        """

    @property
    def body(self):
        if self._body is None:
            canvas = get_canvas()
            for segment in self.segments:
                canvas[segment.body > 0] = 255
            iteration = 1
            while True:
                canvas = close(canvas, iteration=iteration)
                _, num = split(canvas)
                if num == 1:
                    break
                iteration += 1
            self._body = canvas
        return self._body

    @property
    def bone(self):
        if self._bone is None:
            self._bone = skeletonize_(self.body, min_area=-1)
        return self._bone

    @property
    def corner(self):
        if self._corner is None:
            # 获取连续坐标
            start_point = np.transpose(np.nonzero(get_points(self.bone)[1]))[0]
            points = np.transpose(np.nonzero(self.bone))
            curve = np.array(sorted(points, key=lambda x: distance(x, start_point)))[:, ::-1]
            epsilon = 5.0
            approx_curve = cv2.approxPolyDP(curve, epsilon, True)
            angles = []
            for i in range(1, len(approx_curve) - 1):
                p1 = approx_curve[i - 1][0]
                p2 = approx_curve[i][0]
                p3 = approx_curve[i + 1][0]
                angle = calculate_angle(p1, p2, p3)
                angles.append(angle)
            self._corner = angles
        return self._corner

    @property
    def min_corner(self):
        """ 最小拐角 """
        return np.min(self.corner) if self.corner else 180

    @property
    def mean_corner(self):
        return np.mean(self.corner) if self.corner else 180


    @property
    def length(self):
        return np.sum([s.length for s in self.segments])

    @property
    def linearity(self):
        """ 直线度 """
        if self.length:
            return distance(self.nodes[self.start_node].center, self.nodes[self.end_node].center) / self.length
        return 0

    @property
    def area(self):
        return cv2.countNonZero(self.body)



def get_graph(segments, nodes):
    """ 创建无向图 """
    graph = nx.Graph()
    edges = [[n.index for n in s.neighbors] + [s.length, s.index] for s in segments if len(s.neighbors) == 2]
    for u, v, weight, edge_id in edges:
        graph.add_edge(u, v, weight=weight, edge_id=edge_id)

    return graph


def get_shortest(graph, start_node, end_node):
    """ 使用Dijkstra算法计算最短路径 """
    shortest_nodes = nx.dijkstra_path(graph, start_node, end_node)
    shortest_length = nx.dijkstra_path_length(graph, start_node, end_node)
    # 获取最短路径上的边及其编号
    shortest_edges = []
    for i in range(len(shortest_nodes) - 1):
        u, v = shortest_nodes[i], shortest_nodes[i + 1]
        edge_data = graph.get_edge_data(u, v)
        shortest_edges.append(edge_data['edge_id'])


    return shortest_edges, shortest_nodes, shortest_length



def filter_trunk2(trunks, nodes):
    """ 已弃用 """
    m = [x.index for x in nodes]
    subsets = [([n.index for n in t.nodes], t.min_corner * t.area) for t in trunks]
    # 按照x降序排序子集
    sorted_subsets = sorted(subsets, key=lambda x: x[1], reverse=True)

    result = []
    used_elements = set()

    # 选择不重复的子集
    for subset, x in sorted_subsets:
        if all(element not in used_elements for element in subset):
            result.append(subset)
            used_elements.update(subset)

    return result


def filter_trunk(trunks, nodes):
    """ 利用DP找出互不交叉的直线性最高的主干们 """
    m = [x.index for x in nodes]
    subsets = [([n.index for n in t.nodes], t.min_corner * t.area) for t in trunks]
    subset_indices = {tuple(subset): x for subset, x in subsets}
    subset_masks = {i: 0 for i in range(len(subset_indices))}

    for i, (subset, _) in enumerate(subsets):
        subset_masks[i] = sum((1 << m.index(num)) for num in subset)

    @lru_cache(None)
    def dp(mask):
        max_avg = 0
        best_subset = -1
        count = 0
        for i in range(len(subset_masks)):
            if mask & subset_masks[i] == 0:
                current_avg, _, current_count = dp(mask | subset_masks[i])
                current_avg = (current_avg * current_count + subset_indices[tuple(subsets[i][0])]) / (current_count + 1)
                if current_avg > max_avg:
                    max_avg = current_avg
                    best_subset = i
                    count = current_count + 1
        return max_avg, best_subset, count

    max_avg, _, _ = dp(0)

    # Trace back to find the subsets used
    used_subsets = []
    mask = 0
    while True:
        _, best_subset, _ = dp(mask)
        if best_subset == -1:
            break
        used_subsets.append(subsets[best_subset][0])
        mask |= subset_masks[best_subset]

    return used_subsets

def get_trunk(segments, nodes):
    """ 获取主干 """
    trunk_list = []
    graph = get_graph(segments, nodes)
    shortest_dict = {}
    for start_node, end_node in itertools.combinations([n.index for n in nodes if n.class_node == 'end'], 2):
        shortest_edges, shortest_nodes, length = get_shortest(graph, start_node, end_node)
        shortest_dict[tuple(shortest_nodes)] = shortest_edges

        trunk = Trunk(
            NerveContainer([segments[i] for i in shortest_edges]),
            NerveContainer([nodes[i] for i in shortest_nodes]),
            graph,
            start_node,
            end_node
        )

        if trunk.linearity > .8 and trunk.length > 200 and trunk.min_corner > 75:
            trunk_list.append(trunk)

    trunks = filter_trunk(trunk_list, nodes)
    return [(shortest_dict[tuple(t)], t) for t in trunks] if trunks else None



from .utils.morphology import skeletonize_
from .segment.detector import segmenter

import numpy as np


def get_binary(image: np.ndarray) -> np.ndarray:
    return segmenter(image)


def get_skeleton(image: np.ndarray) -> np.ndarray:
    return skeletonize_(image)



import numpy as np
import cv2
import os

from .transforms import Normalize, Compose
from onnxruntime import InferenceSession

CANVAS_SHAPE = 384, 384


class BaseDetector:
    def __init__(self, onnx_path):
        self.sess = InferenceSession(onnx_path)
        self.transform = Compose([Normalize()])

    def infer(self, image):
        if len(image.shape) == 2:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
        inputs = self.transform({'img': image})['img']
        inputs = inputs[np.newaxis, ...]
        inputs = inputs[..., :CANVAS_SHAPE[0], :CANVAS_SHAPE[1]]
        ort_outs = self.sess.run(output_names=None, input_feed={self.sess.get_inputs()[0].name: inputs})
        return ort_outs

    def postprocess(self, ort_outs):
        pass

    def __call__(self, image):
        ort_outs = self.infer(image)
        image = self.postprocess(ort_outs)
        return image


class NerveSegmenter(BaseDetector):
    def __init__(self, onnx_path):
        super().__init__(onnx_path)

    def postprocess(self, ort_outs):
        image = np.squeeze(ort_outs[0]) * 255
        image = image.astype('uint8')
        return image


here = os.path.abspath(os.path.dirname(__file__))


class Segmenter:
    def __init__(self):
        self.segmenter = None
        self.model_path = os.path.join(here, './nerve.onnx')
        try:
            self.set_segmenter(self.model_path)
        except:
            pass

    def set_segmenter(self, model_path):
        self.segmenter = NerveSegmenter(model_path)
        self.model_path = model_path

    def __call__(self, *args, **kwargs):
        return self.segmenter(*args, **kwargs)


segmenter = Segmenter()


def set_segmenter(model_path):
    global segmenter
    segmenter.set_segmenter(model_path)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import numpy as np
from PIL import Image, ImageEnhance
from scipy.ndimage import distance_transform_edt


def crop(img, crop_coordinate):
    x1, y1, x2, y2 = crop_coordinate
    img = img[y1:y2, x1:x2, ...]
    return img


def rescale_size(img_size, target_size):
    scale = min(
        max(target_size) / max(img_size), min(target_size) / min(img_size))
    rescaled_size = [round(i * scale) for i in img_size]
    return rescaled_size, scale


def normalize(im, mean, std):
    im = im.astype(np.float32, copy=False) / 255.0
    im -= mean
    im /= std
    return im


def resize(im, target_size=608, interp=cv2.INTER_LINEAR):
    if isinstance(target_size, list) or isinstance(target_size, tuple):
        w = target_size[0]
        h = target_size[1]
    else:
        w = target_size
        h = target_size
    im = cv2.resize(im, (w, h), interpolation=interp)
    return im


def resize_long(im, long_size=224, interpolation=cv2.INTER_LINEAR):
    value = max(im.shape[0], im.shape[1])
    scale = float(long_size) / float(value)
    resized_width = int(round(im.shape[1] * scale))
    resized_height = int(round(im.shape[0] * scale))

    im = cv2.resize(
        im, (resized_width, resized_height), interpolation=interpolation)
    return im


def resize_short(im, short_size=224, interpolation=cv2.INTER_LINEAR):
    value = min(im.shape[0], im.shape[1])
    scale = float(short_size) / float(value)
    resized_width = int(round(im.shape[1] * scale))
    resized_height = int(round(im.shape[0] * scale))

    im = cv2.resize(
        im, (resized_width, resized_height), interpolation=interpolation)
    return im


def horizontal_flip(im):
    if len(im.shape) == 3:
        im = im[:, ::-1, :]
    elif len(im.shape) == 2:
        im = im[:, ::-1]
    return im


def vertical_flip(im):
    if len(im.shape) == 3:
        im = im[::-1, :, :]
    elif len(im.shape) == 2:
        im = im[::-1, :]
    return im


def brightness(im, brightness_lower, brightness_upper):
    brightness_delta = np.random.uniform(brightness_lower, brightness_upper)
    im = ImageEnhance.Brightness(im).enhance(brightness_delta)
    return im


def contrast(im, contrast_lower, contrast_upper):
    contrast_delta = np.random.uniform(contrast_lower, contrast_upper)
    im = ImageEnhance.Contrast(im).enhance(contrast_delta)
    return im


def saturation(im, saturation_lower, saturation_upper):
    saturation_delta = np.random.uniform(saturation_lower, saturation_upper)
    im = ImageEnhance.Color(im).enhance(saturation_delta)
    return im


def hue(im, hue_lower, hue_upper):
    hue_delta = np.random.uniform(hue_lower, hue_upper)
    im = np.array(im.convert('HSV'))
    im[:, :, 0] = im[:, :, 0] + hue_delta
    im = Image.fromarray(im, mode='HSV').convert('RGB')
    return im


def sharpness(im, sharpness_lower, sharpness_upper):
    sharpness_delta = np.random.uniform(sharpness_lower, sharpness_upper)
    im = ImageEnhance.Sharpness(im).enhance(sharpness_delta)
    return im


def rotate(im, rotate_lower, rotate_upper):
    rotate_delta = np.random.uniform(rotate_lower, rotate_upper)
    im = im.rotate(int(rotate_delta))
    return im


def mask_to_onehot(mask, num_classes):
    """
    Convert a mask (H, W) to onehot (K, H, W).

    Args:
        mask (np.ndarray): Label mask with shape (H, W)
        num_classes (int): Number of classes.

    Returns:
        np.ndarray: Onehot mask with shape(K, H, W).
    """
    _mask = [mask == i for i in range(num_classes)]
    _mask = np.array(_mask).astype(np.uint8)
    return _mask


def onehot_to_binary_edge(mask, radius):
    """
    Convert a onehot mask (K, H, W) to a edge mask.

    Args:
        mask (np.ndarray): Onehot mask with shape (K, H, W)
        radius (int|float): Radius of edge.

    Returns:
        np.ndarray: Edge mask with shape(H, W).
    """
    if radius < 1:
        raise ValueError('`radius` should be greater than or equal to 1')
    num_classes = mask.shape[0]

    edge = np.zeros(mask.shape[1:])
    # pad borders
    mask = np.pad(mask, ((0, 0), (1, 1), (1, 1)),
                  mode='constant',
                  constant_values=0)
    for i in range(num_classes):
        dist = distance_transform_edt(mask[i, :]) + distance_transform_edt(
            1.0 - mask[i, :])
        dist = dist[1:-1, 1:-1]
        dist[dist > radius] = 0
        edge += dist

    edge = np.expand_dims(edge, axis=0)
    edge = (edge > 0).astype(np.uint8)
    return edge


def mask_to_binary_edge(mask, radius, num_classes):
    """
    Convert a segmentic segmentation mask (H, W) to a binary edge mask(H, W).

    Args:
        mask (np.ndarray): Label mask with shape (H, W)
        radius (int|float): Radius of edge.
        num_classes (int): Number of classes.

    Returns:
        np.ndarray: Edge mask with shape(H, W).
    """
    mask = mask.squeeze()
    onehot = mask_to_onehot(mask, num_classes)
    edge = onehot_to_binary_edge(onehot, radius)
    return edge



from setuptools import setup, find_packages


def parse_requirements(filename):
    with open(filename, encoding='utf-8') as f:
        return f.read().splitlines()


setup(
    name='aiccm',
    version='1.1.0',
    author='Qincheng Qiao',
    author_email='jugking6688@gmail.com',
    description='用于分析和处理角膜共聚焦显微镜角膜神经图像的Python软件包',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://https://github.com/SummerColdWind',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ],
    python_requires='>=3.9',
    install_requires=parse_requirements('requirements.txt'),
    include_package_data=True
)



import cv2
import numpy as np

import warnings


def load_image(path):
    """ 加载一张图片 """
    with open(path, 'rb') as file:
        file_data = file.read()
        image_array = np.frombuffer(file_data, np.uint8)
        image_raw = cv2.imdecode(image_array, cv2.IMREAD_COLOR)

    return image_raw


def save_image(image, path='result.png'):
    """ 保存一张图片 """
    extend = '.' + path.split('.')[-1]
    retval, buffer = cv2.imencode(extend, image.astype('uint8'))
    with open(path, 'wb') as f:
        f.write(buffer)


def show_image(image):
    """ 展示一张图片 """
    image_show = image.copy().astype('uint8')
    if np.amax(image_show) == 1:
        image_show = image_show * 255
    cv2.imshow('Show', image_show)
    cv2.waitKey(0)


def load_ccm_image(path):
    """ 加载标准化CCM图像 """
    image_raw = load_image(path)
    match image_raw.shape[:2]:
        case 384, 384:
            return image_raw
        case 484, 384:
            return image_raw[:384, :384]
        case _:
            warnings.warn(
                f"读取的CCM图像分辨率不符合预期，应该为384*384或384*484，而此图像{path}为{image_raw.shape[:2]}",
                UserWarning
            )



# interface
from .utils.io import load_image, save_image, load_ccm_image, show_image
from .functions import get_skeleton, get_binary
from .metrics.interface import get_metrics, get_CNFL, get_CNFD, get_CNBD, get_body_image, get_bone_image

from .segment.detector import segmenter, set_segmenter





import math

import numpy as np

VIEW_LENGTH = 400  # μm
VIEW_PIXEL = 384
CNBD_MIN_LENGTH = 20


def get_view_area_mm2():
    """ 视野面积(mm2) """
    return (VIEW_LENGTH / 1000) ** 2


def pixel_to_mm(num):
    """ 将像素换算成mm2长度 """
    return num * (VIEW_LENGTH / 1000) / VIEW_PIXEL


def get_length(p):
    """ 所有节段长度(像素) """
    return np.sum(s.length for s in p.segments)


def get_nerve_num(p):
    """ 所有神经纤维主干数目 """
    return len(p.nerves)


def get_nerve_branch_num(p):
    """ 所有从神经纤维主干发出的分支的数目 """
    branch_records, branch_count = [], 0
    for nerve in p.nerves:
        for node in [n for n in nerve.nodes if n.class_node == 'branch']:
            branches = [s.index for s in node.neighbors if p.segments[s.index].class_segment == 'branch']
            branches = [si for si in branches if si not in branch_records]
            branches = [si for si in branches if p.segments[si].length > CNBD_MIN_LENGTH]
            branch_count += len(branches)
            branch_records.extend(branches)

    return branch_count

def get_CNFL_(p):
    """ 角膜神经纤维长度（CNFL，mm/mm2）：每平方毫米所有神经纤维长度之和 """
    return pixel_to_mm(get_length(p)) / get_view_area_mm2()


def get_CNFD_(p):
    """ 角膜神经纤维密度（CNFD） 神经总数/ mm2"""
    return get_nerve_num(p) / get_view_area_mm2()


def get_CNBD_(p):
    """ 角膜神经分支密度(CNBD)，主要神经干分支数/mm2 """
    return get_nerve_branch_num(p) / get_view_area_mm2()






from .processor import Processor
from .metrics import get_CNFL_, get_CNFD_, get_CNBD_
from .draw import draw_result_image

import numpy as np

p = Processor()


def get_CNFL(image: np.ndarray) -> float:
    p.load_image(image)
    p.process(trunk_rec=False)
    return round(get_CNFL_(p), 4)

def get_CNFD(image: np.ndarray) -> float:
    p.load_image(image)
    p.process(trunk_rec=True)
    return round(get_CNFD_(p), 4)

def get_CNBD(image: np.ndarray) -> float:
    p.load_image(image)
    p.process(trunk_rec=True)
    return round(get_CNBD_(p), 4)

def get_metrics(image: np.ndarray) -> dict[str, float]:
    p.load_image(image)
    p.process(trunk_rec=True)
    return {
        'CNFL': round(get_CNFL_(p), 4),
        'CNFD': round(get_CNFD_(p), 4),
        'CNBD': round(get_CNBD_(p), 4),
    }

def get_bone_image(image: np.ndarray | None = None) -> np.ndarray:
    return draw_result_image(
        p.segments, p.nodes, image=image, show='bone'
    )

def get_body_image(image: np.ndarray | None = None) -> np.ndarray:
    return draw_result_image(
        p.segments, p.nodes, image=image, show='body'
    )




# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import math

import cv2
import numpy as np
from PIL import Image

from .manager import TRANSFORMS
from .functional import *



@TRANSFORMS.add_component
class Compose:
    """
    Do transformation on input data with corresponding pre-processing and augmentation operations.
    The shape of input data to all operations is [height, width, channels].

    Args:
        transforms (list): A list contains data pre-processing or augmentation. An empty list means only reading images, no transformation.
        to_rgb (bool, optional): If converting image to RGB color space. Default: True.
        img_channels (int, optional): The image channels used to check the loaded image. Default: 3.

    Raises:
        TypeError: When 'transforms' is not a list.
        ValueError: when the length of 'transforms' is less than 1.
    """

    def __init__(self, transforms, to_rgb=True, img_channels=3):
        if not isinstance(transforms, list):
            raise TypeError('The transforms must be a list!')
        self.transforms = transforms
        self.to_rgb = to_rgb
        self.img_channels = img_channels
        self.read_flag = cv2.IMREAD_GRAYSCALE if img_channels == 1 else cv2.IMREAD_COLOR

    def __call__(self, data):
        """
        Args:
            data: A dict to deal with. It may include keys: 'img', 'label', 'trans_info' and 'gt_fields'.
                'trans_info' reserve the image shape informating. And the 'gt_fields' save the key need to transforms
                together with 'img'

        Returns: A dict after process。
        """
        if 'img' not in data.keys():
            raise ValueError("`data` must include `img` key.")
        # data['img'] is numpy array in eg1800 and supervisely
        if data['img'] is None:
            raise TypeError(
                "Expect `data[img]` to be str or np.ndarray, but got NoneType.")
        elif isinstance(data['img'], str):
            img = cv2.imread(data['img'], self.read_flag)
            if img is None:
                raise ValueError('Can\'t read The image file {}!'.format(data['img']))
            data['img'] = img.astype('float32')
        if not isinstance(data['img'], np.ndarray):
            raise TypeError(
                "Expect image to be np.ndarray, but got {}".format(type(data['img'])))

        img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
        if img_channels != self.img_channels:
            raise ValueError(
                'The img_channels ({}) is not equal to the channel of loaded image ({})'.
                format(self.img_channels, img_channels))
        if self.to_rgb and img_channels == 3:
            data['img'] = cv2.cvtColor(data['img'], cv2.COLOR_BGR2RGB)

        if 'label' in data.keys() and isinstance(data['label'], str):
            data['label'] = np.asarray(Image.open(data['label']))
            img_h, img_w = data['img'].shape[:2]
            if data['label'].shape[0] != img_h:
                data['label'] = data['label'].reshape([-1, img_h, img_w]).transpose([1, 2, 0])
            elif data['label'].shape[1] != img_w:
                data['label'] = data['label'].reshape([img_h, -1, img_w]).transpose([0, 2, 1])

        # the `trans_info` will save the process of image shape, and will be used in evaluation and prediction.
        if 'trans_info' not in data.keys():
            data['trans_info'] = []

        for op in self.transforms:
            data = op(data)

        if data['img'].ndim == 2:
            data['img'] = data['img'][..., np.newaxis]
        data['img'] = np.transpose(data['img'], (2, 0, 1))
        if 'label' in data and data['label'].ndim == 3:
            data['label'] = np.transpose(data['label'], (2, 0, 1))
        return data


@TRANSFORMS.add_component
class RandomHorizontalFlip:
    """
    Flip an image horizontally with a certain probability.

    Args:
        prob (float, optional): A probability of horizontally flipping. Default: 0.5.
    """

    def __init__(self, prob=0.5):
        self.prob = prob

    def __call__(self, data):
        if random.random() < self.prob:
            data['img'] = horizontal_flip(data['img'])
            for key in data.get('gt_fields', []):
                data[key] = horizontal_flip(data[key])
        return data


@TRANSFORMS.add_component
class RandomVerticalFlip:
    """
    Flip an image vertically with a certain probability.

    Args:
        prob (float, optional): A probability of vertical flipping. Default: 0.1.
    """

    def __init__(self, prob=0.1):
        self.prob = prob

    def __call__(self, data):
        if random.random() < self.prob:
            data['img'] = vertical_flip(data['img'])
            for key in data.get('gt_fields', []):
                data[key] = vertical_flip(data[key])
        return data


@TRANSFORMS.add_component
class Resize:
    """
    Resize an image.

    Args:
        target_size (list|tuple, optional): The target size (w, h) of image. Default: (512, 512).
        keep_ratio (bool, optional): Whether to keep the same ratio for width and height in resizing.
            Default: False.
        size_divisor (int, optional): If size_divisor is not None, make the width and height be the times
            of size_divisor. Default: None.
        interp (str, optional): The interpolation mode of resize is consistent with opencv.
            ['NEAREST', 'LINEAR', 'CUBIC', 'AREA', 'LANCZOS4', 'RANDOM']. Note that when it is
            'RANDOM', a random interpolation mode would be specified. Default: "LINEAR".

    Raises:
        TypeError: When 'target_size' type is neither list nor tuple.
        ValueError: When "interp" is out of pre-defined methods ('NEAREST', 'LINEAR', 'CUBIC',
        'AREA', 'LANCZOS4', 'RANDOM').
    """

    # The interpolation mode
    interp_dict = {
        'NEAREST': cv2.INTER_NEAREST,
        'LINEAR': cv2.INTER_LINEAR,
        'CUBIC': cv2.INTER_CUBIC,
        'AREA': cv2.INTER_AREA,
        'LANCZOS4': cv2.INTER_LANCZOS4
    }

    def __init__(self,
                 target_size=(512, 512),
                 keep_ratio=False,
                 size_divisor=None,
                 interp='LINEAR'):
        if isinstance(target_size, list) or isinstance(target_size, tuple):
            if len(target_size) != 2:
                raise ValueError(
                    '`target_size` should include 2 elements, but it is {}'.
                    format(target_size))
        else:
            raise TypeError(
                "Type of `target_size` is invalid. It should be list or tuple, but it is {}"
                .format(type(target_size)))
        if not (interp == "RANDOM" or interp in self.interp_dict):
            raise ValueError("`interp` should be one of {}".format(
                self.interp_dict.keys()))
        if size_divisor is not None:
            assert isinstance(size_divisor,
                              int), "size_divisor should be None or int"

        self.target_size = target_size
        self.keep_ratio = keep_ratio
        self.size_divisor = size_divisor
        self.interp = interp

    def __call__(self, data):
        data['trans_info'].append(('resize', data['img'].shape[0:2]))
        if self.interp == "RANDOM":
            interp = random.choice(list(self.interp_dict.keys()))
        else:
            interp = self.interp

        target_size = self.target_size
        if self.keep_ratio:
            h, w = data['img'].shape[0:2]
            target_size, _ = rescale_size((w, h), self.target_size)
        if self.size_divisor:
            target_size = [
                math.ceil(i / self.size_divisor) * self.size_divisor
                for i in target_size
            ]

        data['img'] = resize(data['img'], target_size,
                                        self.interp_dict[interp])
        for key in data.get('gt_fields', []):
            data[key] = resize(data[key], target_size,
                                          cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class ResizeByLong:
    """
    Resize the long side of an image to given size, and then scale the other side proportionally.

    Args:
        long_size (int): The target size of long side.
    """

    def __init__(self, long_size):
        self.long_size = long_size

    def __call__(self, data):
        data['trans_info'].append(('resize', data['img'].shape[0:2]))
        data['img'] = resize_long(data['img'], self.long_size)
        for key in data.get('gt_fields', []):
            data[key] = resize_long(data[key], self.long_size,
                                               cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class ResizeByShort:
    """
    Resize the short side of an image to given size, and then scale the other side proportionally.

    Args:
        short_size (int): The target size of short side.
        max_size(int): The maximum length of resized image's long edge, if the resized image's long edge exceed this length, short size will be adjusted.
    """

    def __init__(self, short_size, max_size=1e10):
        if isinstance(short_size, list):
            short_size = random.choice(short_size)
        self.short_size = short_size
        self.max_size = max_size

    def __call__(self, data):
        h, w = data['img'].shape[0:2]
        data['trans_info'].append(('resize', data['img'].shape[0:2]))
        if self.short_size / min(h, w) * max(h, w) > self.max_size:
            self.short_size = int((self.max_size / max(h, w)) * min(h, w))

        data['img'] = resize_short(data['img'], self.short_size)
        for key in data.get('gt_fields', []):
            data[key] = resize_short(data[key], self.short_size,
                                                cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class LimitLong:
    """
    Limit the long edge of image.

    If the long edge is larger than max_long, resize the long edge
    to max_long, while scale the short edge proportionally.

    If the long edge is smaller than min_long, resize the long edge
    to min_long, while scale the short edge proportionally.

    Args:
        max_long (int, optional): If the long edge of image is larger than max_long,
            it will be resize to max_long. Default: None.
        min_long (int, optional): If the long edge of image is smaller than min_long,
            it will be resize to min_long. Default: None.
    """

    def __init__(self, max_long=None, min_long=None):
        if max_long is not None:
            if not isinstance(max_long, int):
                raise TypeError(
                    "Type of `max_long` is invalid. It should be int, but it is {}"
                    .format(type(max_long)))
        if min_long is not None:
            if not isinstance(min_long, int):
                raise TypeError(
                    "Type of `min_long` is invalid. It should be int, but it is {}"
                    .format(type(min_long)))
        if (max_long is not None) and (min_long is not None):
            if min_long > max_long:
                raise ValueError(
                    '`max_long should not smaller than min_long, but they are {} and {}'
                    .format(max_long, min_long))
        self.max_long = max_long
        self.min_long = min_long

    def __call__(self, data):
        data['trans_info'].append(('resize', data['img'].shape[0:2]))

        h, w = data['img'].shape[0], data['img'].shape[1]
        long_edge = max(h, w)
        target = long_edge
        if (self.max_long is not None) and (long_edge > self.max_long):
            target = self.max_long
        elif (self.min_long is not None) and (long_edge < self.min_long):
            target = self.min_long

        if target != long_edge:
            data['img'] = resize_long(data['img'], target)
            for key in data.get('gt_fields', []):
                data[key] = resize_long(data[key], target,
                                                   cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class ResizeRangeScaling:
    """
    Resize the long side of an image into a range, and then scale the other side proportionally.

    Args:
        min_value (int, optional): The minimum value of long side after resize. Default: 400.
        max_value (int, optional): The maximum value of long side after resize. Default: 600.
    """

    def __init__(self, min_value=400, max_value=600):
        if min_value > max_value:
            raise ValueError('min_value must be less than max_value, '
                             'but they are {} and {}.'.format(min_value,
                                                              max_value))
        self.min_value = min_value
        self.max_value = max_value

    def __call__(self, data):

        if self.min_value == self.max_value:
            random_size = self.max_value
        else:
            random_size = int(
                np.random.uniform(self.min_value, self.max_value) + 0.5)
        data['img'] = resize_long(data['img'], random_size,
                                             cv2.INTER_LINEAR)
        for key in data.get('gt_fields', []):
            data[key] = resize_long(data[key], random_size,
                                               cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class ResizeStepScaling:
    """
    Scale an image proportionally within a range.

    Args:
        min_scale_factor (float, optional): The minimum scale. Default: 0.75.
        max_scale_factor (float, optional): The maximum scale. Default: 1.25.
        scale_step_size (float, optional): The scale interval. Default: 0.25.

    Raises:
        ValueError: When min_scale_factor is smaller than max_scale_factor.
    """

    def __init__(self,
                 min_scale_factor=0.75,
                 max_scale_factor=1.25,
                 scale_step_size=0.25):
        if min_scale_factor > max_scale_factor:
            raise ValueError(
                'min_scale_factor must be less than max_scale_factor, '
                'but they are {} and {}.'.format(min_scale_factor,
                                                 max_scale_factor))
        self.min_scale_factor = min_scale_factor
        self.max_scale_factor = max_scale_factor
        self.scale_step_size = scale_step_size

    def __call__(self, data):

        if self.min_scale_factor == self.max_scale_factor:
            scale_factor = self.min_scale_factor

        elif self.scale_step_size == 0:
            scale_factor = np.random.uniform(self.min_scale_factor,
                                             self.max_scale_factor)

        else:
            num_steps = int((self.max_scale_factor - self.min_scale_factor) /
                            self.scale_step_size + 1)
            scale_factors = np.linspace(self.min_scale_factor,
                                        self.max_scale_factor,
                                        num_steps).tolist()
            np.random.shuffle(scale_factors)
            scale_factor = scale_factors[0]
        w = int(round(scale_factor * data['img'].shape[1]))
        h = int(round(scale_factor * data['img'].shape[0]))

        data['img'] = resize(data['img'], (w, h), cv2.INTER_LINEAR)
        for key in data.get('gt_fields', []):
            data[key] = resize(data[key], (w, h), cv2.INTER_NEAREST)

        return data


@TRANSFORMS.add_component
class Normalize:
    """
    Normalize an image.

    Args:
        mean (list, optional): The mean value of a data set. Default: [0.5,].
        std (list, optional): The standard deviation of a data set. Default: [0.5,].

    Raises:
        ValueError: When mean/std is not list or any value in std is 0.
    """

    def __init__(self, mean=(0.5, ), std=(0.5, )):
        if not (isinstance(mean, (list, tuple)) and isinstance(std, (list, tuple))) \
            and (len(mean) not in [1, 3]) and (len(std) not in [1, 3]):
            raise ValueError(
                "{}: input type is invalid. It should be list or tuple with the lenght of 1 or 3".
                format(self))
        self.mean = np.array(mean)
        self.std = np.array(std)

        from functools import reduce
        if reduce(lambda x, y: x * y, self.std) == 0:
            raise ValueError('{}: std is invalid!'.format(self))

    def __call__(self, data):
        data['img'] = normalize(data['img'], self.mean, self.std)
        return data


@TRANSFORMS.add_component
class Padding:
    """
    Add bottom-right padding to a raw image or annotation image.

    Args:
        target_size (list|tuple): The target size after padding.
        im_padding_value (float, optional): The padding value of raw image.
            Default: 127.5.
        label_padding_value (int, optional): The padding value of annotation image. Default: 255.

    Raises:
        TypeError: When target_size is neither list nor tuple.
        ValueError: When the length of target_size is not 2.
    """

    def __init__(self,
                 target_size,
                 im_padding_value=127.5,
                 label_padding_value=255):
        if isinstance(target_size, list) or isinstance(target_size, tuple):
            if len(target_size) != 2:
                raise ValueError(
                    '`target_size` should include 2 elements, but it is {}'.
                    format(target_size))
        else:
            raise TypeError(
                "Type of target_size is invalid. It should be list or tuple, now is {}"
                .format(type(target_size)))
        self.target_size = target_size
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value

    def __call__(self, data):
        data['trans_info'].append(('padding', data['img'].shape[0:2]))
        im_height, im_width = data['img'].shape[0], data['img'].shape[1]
        if isinstance(self.target_size, int):
            target_height = self.target_size
            target_width = self.target_size
        else:
            target_height = self.target_size[1]
            target_width = self.target_size[0]
        pad_height = target_height - im_height
        pad_width = target_width - im_width
        if pad_height < 0 or pad_width < 0:
            raise ValueError(
                'The size of image should be less than `target_size`, but the size of image ({}, {}) is larger than `target_size` ({}, {})'
                .format(im_width, im_height, target_width, target_height))
        else:
            img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
            data['img'] = cv2.copyMakeBorder(
                data['img'],
                0,
                pad_height,
                0,
                pad_width,
                cv2.BORDER_CONSTANT,
                value=(self.im_padding_value, ) * img_channels)
            for key in data.get('gt_fields', []):
                data[key] = cv2.copyMakeBorder(
                    data[key],
                    0,
                    pad_height,
                    0,
                    pad_width,
                    cv2.BORDER_CONSTANT,
                    value=self.label_padding_value)
        return data


@TRANSFORMS.add_component
class PaddingByAspectRatio:
    """

    Args:
        aspect_ratio (int|float, optional): The aspect ratio = width / height. Default: 1.
        im_padding_value (float, optional): The padding value of raw image. Default: 127.5.
        label_padding_value (int, optional): The padding value of annotation image. Default: 255.
    """

    def __init__(self,
                 aspect_ratio=1,
                 im_padding_value=127.5,
                 label_padding_value=255):
        self.aspect_ratio = aspect_ratio
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value

    def __call__(self, data):

        img_height = data['img'].shape[0]
        img_width = data['img'].shape[1]
        ratio = img_width / img_height
        if ratio == self.aspect_ratio:
            return data
        elif ratio > self.aspect_ratio:
            img_height = int(img_width / self.aspect_ratio)
        else:
            img_width = int(img_height * self.aspect_ratio)
        padding = Padding(
            (img_width, img_height),
            im_padding_value=self.im_padding_value,
            label_padding_value=self.label_padding_value)
        return padding(data)


@TRANSFORMS.add_component
class RandomPaddingCrop:
    """
    Crop a sub-image from a raw image and annotation image randomly. If the target cropping size
    is larger than original image, then the bottom-right padding will be added.

    Args:
        crop_size (tuple, optional): The target cropping size. Default: (512, 512).
        im_padding_value (float, optional): The padding value of raw image. Default: 127.5.
        label_padding_value (int, optional): The padding value of annotation image. Default: 255.
        category_max_ratio (float, optional): The maximum ratio that single category could occupy. 
            Default: 1.0.
        ignore_index (int, optional): The value that should be ignored in the annotation image. 
            Default: 255.
        loop_times (int, optional): The maximum number of attempts to crop an image. Default: 10.

    Raises:
        TypeError: When crop_size is neither list nor tuple.
        ValueError: When the length of crop_size is not 2.
    """

    def __init__(self,
                 crop_size=(512, 512),
                 im_padding_value=127.5,
                 label_padding_value=255,
                 category_max_ratio=1.0,
                 ignore_index=255,
                 loop_times=10):
        if isinstance(crop_size, list) or isinstance(crop_size, tuple):
            if len(crop_size) != 2:
                raise ValueError(
                    'Type of `crop_size` is list or tuple. It should include 2 elements, but it is {}.'
                    .format(crop_size))
        else:
            raise TypeError(
                "The type of `crop_size` is invalid. It should be list or tuple, but it is {}."
                .format(type(crop_size)))
        if category_max_ratio <= 0:
            raise ValueError(
                "The value of `category_max_ratio` must be greater than 0, but got {}.".
                format(category_max_ratio))
        if loop_times <= 0:
            raise ValueError(
                "The value of `loop_times` must be greater than 0, but got {}.".
                format(loop_times))

        self.crop_size = tuple(reversed(crop_size))
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value
        self.category_max_ratio = category_max_ratio
        self.ignore_index = ignore_index
        self.loop_times = loop_times

    def _get_crop_coordinates(self, origin_size):
        margin_h = max(origin_size[0] - self.crop_size[0], 0)
        margin_w = max(origin_size[1] - self.crop_size[1], 0)
        offset_h = np.random.randint(0, margin_h + 1)
        offset_w = np.random.randint(0, margin_w + 1)
        crop_y1, crop_y2 = offset_h, offset_h + self.crop_size[0]
        crop_x1, crop_x2 = offset_w, offset_w + self.crop_size[1]

        return crop_x1, crop_y1, crop_x2, crop_y2

    def _padding(self, data):
        img_shape = data['img'].shape[:2]
        pad_height = max(self.crop_size[0] - img_shape[0], 0)
        pad_width = max(self.crop_size[1] - img_shape[1], 0)
        img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
        if (pad_height > 0 or pad_width > 0):
            data['img'] = cv2.copyMakeBorder(
                data['img'],
                0,
                pad_height,
                0,
                pad_width,
                cv2.BORDER_CONSTANT,
                value=(self.im_padding_value, ) * img_channels)
            for key in data.get('gt_fields', []):
                data[key] = cv2.copyMakeBorder(
                    data[key],
                    0,
                    pad_height,
                    0,
                    pad_width,
                    cv2.BORDER_CONSTANT,
                    value=self.label_padding_value)
        return data

    def __call__(self, data):
        img_shape = data['img'].shape[:2]
        if img_shape[0] == self.crop_size[0] and img_shape[1] == self.crop_size[
                1]:
            return data

        data = self._padding(data)
        img_shape = data['img'].shape[:2]
        crop_coordinates = self._get_crop_coordinates(img_shape)

        if self.category_max_ratio < 1.0:
            for _ in range(self.loop_times):
                seg_temp = crop(data["label"], crop_coordinates)
                labels, cnt = np.unique(seg_temp, return_counts=True)
                cnt = cnt[labels != self.ignore_index]
                if len(cnt) > 1 and np.max(cnt) / np.sum(
                        cnt) < self.category_max_ratio:
                    data['img'] = seg_temp
                    break
                crop_coordinates = self._get_crop_coordinates(img_shape)
        else:
            data['img'] = crop(data['img'], crop_coordinates)
        for key in data.get("gt_fields", []):
            data[key] = crop(data[key], crop_coordinates)

        return data


@TRANSFORMS.add_component
class RandomCenterCrop:
    """
    Crops the given the input data at the center.
    Args:
        retain_ratio (tuple or list, optional): The length of the input list or tuple must be 2. Default: (0.5, 0.5).
        the first value is used for width and the second is for height.
        In addition, the minimum size of the cropped image is [width * retain_ratio[0], height * retain_ratio[1]].
    Raises:
        TypeError: When retain_ratio is neither list nor tuple. Default: None.
        ValueError: When the value of retain_ratio is not in [0-1].
    """

    def __init__(self, retain_ratio=(0.5, 0.5)):
        if isinstance(retain_ratio, list) or isinstance(retain_ratio, tuple):
            if len(retain_ratio) != 2:
                raise ValueError(
                    'When type of `retain_ratio` is list or tuple, it shoule include 2 elements, but it is {}'
                    .format(retain_ratio))
            if retain_ratio[0] > 1 or retain_ratio[1] > 1 or retain_ratio[
                    0] < 0 or retain_ratio[1] < 0:
                raise ValueError(
                    'Value of `retain_ratio` should be in [0, 1], but it is {}'.
                    format(retain_ratio))
        else:
            raise TypeError(
                "The type of `retain_ratio` is invalid. It should be list or tuple, but it is {}"
                .format(type(retain_ratio)))
        self.retain_ratio = retain_ratio

    def __call__(self, data):
        retain_width = self.retain_ratio[0]
        retain_height = self.retain_ratio[1]

        img_height = data['img'].shape[0]
        img_width = data['img'].shape[1]

        if retain_width == 1. and retain_height == 1.:
            return data
        else:
            randw = np.random.randint(img_width * (1 - retain_width))
            randh = np.random.randint(img_height * (1 - retain_height))
            offsetw = 0 if randw == 0 else np.random.randint(randw)
            offseth = 0 if randh == 0 else np.random.randint(randh)
            p0, p1, p2, p3 = offseth, img_height + offseth - randh, offsetw, img_width + offsetw - randw
            if data['img'].ndim == 2:
                data['img'] = data['img'][p0:p1, p2:p3]
            else:
                data['img'] = data['img'][p0:p1, p2:p3, :]
            for key in data.get('gt_fields', []):
                data[key] = data[key][p0:p1, p2:p3]

        return data


@TRANSFORMS.add_component
class ScalePadding:
    """
        Add center padding to a raw image or annotation image,then scale the
        image to target size.

        Args:
            target_size (list|tuple, optional): The target size of image. Default: (512, 512).
            im_padding_value (float, optional): The padding value of raw image. Default: 127.5
            label_padding_value (int, optional): The padding value of annotation image. Default: 255.

        Raises:
            TypeError: When target_size is neither list nor tuple.
            ValueError: When the length of target_size is not 2.
    """

    def __init__(self,
                 target_size=(512, 512),
                 im_padding_value=127.5,
                 label_padding_value=255):
        if isinstance(target_size, list) or isinstance(target_size, tuple):
            if len(target_size) != 2:
                raise ValueError(
                    '`target_size` should include 2 elements, but it is {}'.
                    format(target_size))
        else:
            raise TypeError(
                "Type of `target_size` is invalid. It should be list or tuple, but it is {}"
                .format(type(target_size)))

        self.target_size = target_size
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value

    def __call__(self, data):
        height = data['img'].shape[0]
        width = data['img'].shape[1]

        img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
        if data['img'].ndim == 2:
            new_im = np.zeros((max(height, width), max(height, width)
                               )) + self.im_padding_value
        else:
            new_im = np.zeros((max(height, width), max(height, width),
                               img_channels)) + self.im_padding_value
        if 'label' in data['gt_fields']:
            new_label = np.zeros((max(height, width), max(height, width)
                                  )) + self.label_padding_value

        if height > width:
            padding = int((height - width) / 2)
            if data['img'].ndim == 2:
                new_im[:, padding:padding + width] = data['img']
            else:
                new_im[:, padding:padding + width, :] = data['img']
            if 'label' in data['gt_fields']:
                new_label[:, padding:padding + width] = data['label']
        else:
            padding = int((width - height) / 2)
            new_im[padding:padding + height, :] = data['img']
            if 'label' in data['gt_fields']:
                new_label[padding:padding + height, :] = data['label']

        data['img'] = np.uint8(new_im)
        data['img'] = resize(
            data['img'], self.target_size, interp=cv2.INTER_CUBIC)
        if 'label' in data['gt_fields']:
            data['label'] = np.uint8(new_label)
            data['label'] = resize(
                data['label'], self.target_size, interp=cv2.INTER_CUBIC)
        return data


@TRANSFORMS.add_component
class RandomNoise:
    """
    Superimposing noise on an image with a certain probability.

    Args:
        prob (float, optional): A probability of blurring an image. Default: 0.5.
        max_sigma(float, optional): The maximum value of standard deviation of the distribution.
            Default: 10.0.
    """

    def __init__(self, prob=0.5, max_sigma=10.0):
        self.prob = prob
        self.max_sigma = max_sigma

    def __call__(self, data):
        if random.random() < self.prob:
            mu = 0
            sigma = random.random() * self.max_sigma
            data['img'] = np.array(data['img'], dtype=np.float32)
            data['img'] += np.random.normal(mu, sigma, data['img'].shape)
            data['img'][data['img'] > 255] = 255
            data['img'][data['img'] < 0] = 0

        return data


@TRANSFORMS.add_component
class RandomBlur:
    """
    Blurring an image by a Gaussian function with a certain probability.

    Args:
        prob (float, optional): A probability of blurring an image. Default: 0.1.
        blur_type(str, optional): A type of blurring an image,
            gaussian stands for cv2.GaussianBlur,
            median stands for cv2.medianBlur,
            blur stands for cv2.blur,
            random represents randomly selected from above.
            Default: gaussian.
    """

    def __init__(self, prob=0.1, blur_type="gaussian"):
        self.prob = prob
        self.blur_type = blur_type

    def __call__(self, data):

        if self.prob <= 0:
            n = 0
        elif self.prob >= 1:
            n = 1
        else:
            n = int(1.0 / self.prob)
        if n > 0:
            if np.random.randint(0, n) == 0:
                radius = np.random.randint(3, 10)
                if radius % 2 != 1:
                    radius = radius + 1
                if radius > 9:
                    radius = 9
                data['img'] = np.array(data['img'], dtype='uint8')
                if self.blur_type == "gaussian":
                    data['img'] = cv2.GaussianBlur(data['img'],
                                                   (radius, radius), 0, 0)
                elif self.blur_type == "median":
                    data['img'] = cv2.medianBlur(data['img'], radius)
                elif self.blur_type == "blur":
                    data['img'] = cv2.blur(data['img'], (radius, radius))
                elif self.blur_type == "random":
                    select = random.random()
                    if select < 0.3:
                        data['img'] = cv2.GaussianBlur(data['img'],
                                                       (radius, radius), 0)
                    elif select < 0.6:
                        data['img'] = cv2.medianBlur(data['img'], radius)
                    else:
                        data['img'] = cv2.blur(data['img'], (radius, radius))
                else:
                    data['img'] = cv2.GaussianBlur(data['img'],
                                                   (radius, radius), 0, 0)
        data['img'] = np.array(data['img'], dtype='float32')
        return data


@TRANSFORMS.add_component
class RandomRotation:
    """
    Rotate an image randomly with padding.

    Args:
        max_rotation (float, optional): The maximum rotation degree. Default: 15.
        im_padding_value (float, optional): The padding value of raw image. Default: 127.5.
        label_padding_value (int, optional): The padding value of annotation image. Default: 255.
    """

    def __init__(self,
                 max_rotation=15,
                 im_padding_value=127.5,
                 label_padding_value=255):
        self.max_rotation = max_rotation
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value

    def __call__(self, data):

        if self.max_rotation > 0:
            (h, w) = data['img'].shape[:2]
            img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
            do_rotation = np.random.uniform(-self.max_rotation,
                                            self.max_rotation)
            pc = (w // 2, h // 2)
            r = cv2.getRotationMatrix2D(pc, do_rotation, 1.0)
            cos = np.abs(r[0, 0])
            sin = np.abs(r[0, 1])

            nw = int((h * sin) + (w * cos))
            nh = int((h * cos) + (w * sin))

            (cx, cy) = pc
            r[0, 2] += (nw / 2) - cx
            r[1, 2] += (nh / 2) - cy
            dsize = (nw, nh)
            data['img'] = cv2.warpAffine(
                data['img'],
                r,
                dsize=dsize,
                flags=cv2.INTER_LINEAR,
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=(self.im_padding_value, ) * img_channels)
            for key in data.get('gt_fields', []):
                data[key] = cv2.warpAffine(
                    data[key],
                    r,
                    dsize=dsize,
                    flags=cv2.INTER_NEAREST,
                    borderMode=cv2.BORDER_CONSTANT,
                    borderValue=self.label_padding_value)

        return data


@TRANSFORMS.add_component
class RandomScaleAspect:
    """
    Crop a sub-image from an original image with a range of area ratio and aspect and
    then scale the sub-image back to the size of the original image.

    Args:
        min_scale (float, optional): The minimum area ratio of cropped image to the original image. Default: 0.5.
        aspect_ratio (float, optional): The minimum aspect ratio. Default: 0.33.
    """

    def __init__(self, min_scale=0.5, aspect_ratio=0.33):
        self.min_scale = min_scale
        self.aspect_ratio = aspect_ratio

    def __call__(self, data):

        if self.min_scale != 0 and self.aspect_ratio != 0:
            img_height = data['img'].shape[0]
            img_width = data['img'].shape[1]
            for i in range(0, 10):
                area = img_height * img_width
                target_area = area * np.random.uniform(self.min_scale, 1.0)
                aspectRatio = np.random.uniform(self.aspect_ratio,
                                                1.0 / self.aspect_ratio)

                dw = int(np.sqrt(target_area * 1.0 * aspectRatio))
                dh = int(np.sqrt(target_area * 1.0 / aspectRatio))
                if (np.random.randint(10) < 5):
                    tmp = dw
                    dw = dh
                    dh = tmp

                if (dh < img_height and dw < img_width):
                    h1 = np.random.randint(0, img_height - dh)
                    w1 = np.random.randint(0, img_width - dw)

                    if data['img'].ndim == 2:
                        data['img'] = data['img'][h1:(h1 + dh), w1:(w1 + dw)]
                    else:
                        data['img'] = data['img'][h1:(h1 + dh), w1:(w1 + dw), :]
                    data['img'] = cv2.resize(
                        data['img'], (img_width, img_height),
                        interpolation=cv2.INTER_LINEAR)
                    for key in data.get('gt_fields', []):
                        data[key] = data[key][h1:(h1 + dh), w1:(w1 + dw)]
                        data[key] = cv2.resize(
                            data[key], (img_width, img_height),
                            interpolation=cv2.INTER_NEAREST)
                    break
        return data


@TRANSFORMS.add_component
class RandomDistort:
    """
    Distort an image with random configurations.

    Args:
        brightness_range (float, optional): A range of brightness. Default: 0.5.
        brightness_prob (float, optional): A probability of adjusting brightness. Default: 0.5.
        contrast_range (float, optional): A range of contrast. Default: 0.5.
        contrast_prob (float, optional): A probability of adjusting contrast. Default: 0.5.
        saturation_range (float, optional): A range of saturation. Default: 0.5.
        saturation_prob (float, optional): A probability of adjusting saturation. Default: 0.5.
        hue_range (int, optional): A range of hue. Default: 18.
        hue_prob (float, optional): A probability of adjusting hue. Default: 0.5.
        sharpness_range (float, optional): A range of sharpness. Default: 0.5.
        sharpness_prob (float, optional): A probability of adjusting saturation. Default: 0.
    """

    def __init__(self,
                 brightness_range=0.5,
                 brightness_prob=0.5,
                 contrast_range=0.5,
                 contrast_prob=0.5,
                 saturation_range=0.5,
                 saturation_prob=0.5,
                 hue_range=18,
                 hue_prob=0.5,
                 sharpness_range=0.5,
                 sharpness_prob=0):
        self.brightness_range = brightness_range
        self.brightness_prob = brightness_prob
        self.contrast_range = contrast_range
        self.contrast_prob = contrast_prob
        self.saturation_range = saturation_range
        self.saturation_prob = saturation_prob
        self.hue_range = hue_range
        self.hue_prob = hue_prob
        self.sharpness_range = sharpness_range
        self.sharpness_prob = sharpness_prob

    def __call__(self, data):

        brightness_lower = 1 - self.brightness_range
        brightness_upper = 1 + self.brightness_range
        contrast_lower = 1 - self.contrast_range
        contrast_upper = 1 + self.contrast_range
        saturation_lower = 1 - self.saturation_range
        saturation_upper = 1 + self.saturation_range
        hue_lower = -self.hue_range
        hue_upper = self.hue_range
        sharpness_lower = 1 - self.sharpness_range
        sharpness_upper = 1 + self.sharpness_range
        ops = [
            brightness, contrast, saturation,
            sharpness
        ]
        if data['img'].ndim > 2:
            ops.append(hue)
        random.shuffle(ops)
        params_dict = {
            'brightness': {
                'brightness_lower': brightness_lower,
                'brightness_upper': brightness_upper
            },
            'contrast': {
                'contrast_lower': contrast_lower,
                'contrast_upper': contrast_upper
            },
            'saturation': {
                'saturation_lower': saturation_lower,
                'saturation_upper': saturation_upper
            },
            'hue': {
                'hue_lower': hue_lower,
                'hue_upper': hue_upper
            },
            'sharpness': {
                'sharpness_lower': sharpness_lower,
                'sharpness_upper': sharpness_upper,
            }
        }
        prob_dict = {
            'brightness': self.brightness_prob,
            'contrast': self.contrast_prob,
            'saturation': self.saturation_prob,
            'hue': self.hue_prob,
            'sharpness': self.sharpness_prob
        }
        data['img'] = data['img'].astype('uint8')
        data['img'] = Image.fromarray(data['img'])
        for id in range(len(ops)):
            params = params_dict[ops[id].__name__]
            prob = prob_dict[ops[id].__name__]
            params['im'] = data['img']
            if np.random.uniform(0, 1) < prob:
                data['img'] = ops[id](**params)
        data['img'] = np.asarray(data['img']).astype('float32')
        return data


@TRANSFORMS.add_component
class RandomAffine:
    """
    Affine transform an image with random configurations.

    Args:
        size (tuple, optional): The target size after affine transformation. Default: (224, 224).
        translation_offset (float, optional): The maximum translation offset. Default: 0.
        max_rotation (float, optional): The maximum rotation degree. Default: 15.
        min_scale_factor (float, optional): The minimum scale. Default: 0.75.
        max_scale_factor (float, optional): The maximum scale. Default: 1.25.
        im_padding_value (float, optional): The padding value of raw image. Default: 128.
        label_padding_value (int, optional): The padding value of annotation image. Default: (255, 255, 255).
    """

    def __init__(self,
                 size=(224, 224),
                 translation_offset=0,
                 max_rotation=15,
                 min_scale_factor=0.75,
                 max_scale_factor=1.25,
                 im_padding_value=128,
                 label_padding_value=255):
        self.size = size
        self.translation_offset = translation_offset
        self.max_rotation = max_rotation
        self.min_scale_factor = min_scale_factor
        self.max_scale_factor = max_scale_factor
        self.im_padding_value = im_padding_value
        self.label_padding_value = label_padding_value

    def __call__(self, data):

        w, h = self.size
        bbox = [0, 0, data['img'].shape[1] - 1, data['img'].shape[0] - 1]
        x_offset = (random.random() - 0.5) * 2 * self.translation_offset
        y_offset = (random.random() - 0.5) * 2 * self.translation_offset
        dx = (w - (bbox[2] + bbox[0])) / 2.0
        dy = (h - (bbox[3] + bbox[1])) / 2.0

        matrix_trans = np.array([[1.0, 0, dx], [0, 1.0, dy], [0, 0, 1.0]])

        angle = random.random() * 2 * self.max_rotation - self.max_rotation
        scale = random.random() * (self.max_scale_factor - self.min_scale_factor
                                   ) + self.min_scale_factor
        scale *= np.mean(
            [float(w) / (bbox[2] - bbox[0]), float(h) / (bbox[3] - bbox[1])])
        alpha = scale * math.cos(angle / 180.0 * math.pi)
        beta = scale * math.sin(angle / 180.0 * math.pi)

        centerx = w / 2.0 + x_offset
        centery = h / 2.0 + y_offset
        matrix = np.array(
            [[alpha, beta, (1 - alpha) * centerx - beta * centery],
             [-beta, alpha, beta * centerx + (1 - alpha) * centery],
             [0, 0, 1.0]])

        matrix = matrix.dot(matrix_trans)[0:2, :]
        img_channels = 1 if data['img'].ndim == 2 else data['img'].shape[2]
        data['img'] = cv2.warpAffine(
            np.uint8(data['img']),
            matrix,
            tuple(self.size),
            flags=cv2.INTER_LINEAR,
            borderMode=cv2.BORDER_CONSTANT,
            borderValue=(self.im_padding_value, ) * img_channels)
        for key in data.get('gt_fields', []):
            data[key] = cv2.warpAffine(
                np.uint8(data[key]),
                matrix,
                tuple(self.size),
                flags=cv2.INTER_NEAREST,
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=self.label_padding_value)
        return data


@TRANSFORMS.add_component
class GenerateInstanceTargets:
    """
    Generate instance targets from ground-truth labels.

    Args:
        num_classes (int): The number of classes.
        ignore_index (int, optional): Specifies a target value that is ignored. Default: 255.
    """

    def __init__(self, num_classes, ignore_index=255):
        self.num_classes = num_classes
        self.ignore_index = ignore_index

    def __call__(self, data):
        if 'label' in data:
            sem_seg_gt = data['label']
            instances = {"image_shape": data['img'].shape[1:]}
            classes = np.unique(sem_seg_gt)
            classes = classes[classes != self.ignore_index]

            # To make data compatible with dataloader
            classes_cpt = np.array([
                self.ignore_index
                for _ in range(self.num_classes - len(classes))
            ])
            classes_cpt = np.append(classes, classes_cpt)
            instances["gt_classes"] = np.asarray(classes_cpt).astype('int64')

            masks = []
            for cid in classes:
                masks.append(sem_seg_gt == cid)  # [C, H, W] 

            shape = [self.num_classes - len(masks)] + list(data['label'].shape)
            masks_cpt = np.zeros(shape, dtype='int64')

            if len(masks) == 0:
                # Some images do not have annotation and will all be ignored
                instances['gt_masks'] = np.zeros(
                    (self.num_classes, sem_seg_gt.shape[-2],
                     sem_seg_gt.shape[-1]),
                    dtype='int64')

            else:
                instances['gt_masks'] = np.concatenate(
                    [
                        np.stack([
                            np.ascontiguousarray(x).astype('float32')
                            for x in masks
                        ]), masks_cpt
                    ],
                    axis=0)

            data['instances'] = instances

        return data


@TRANSFORMS.add_component
class AddMultiLabelAuxiliaryCategory:
    """
    Add a complementary set of unions labeled with corresponding mask for other categories as an auxiliary category.
    """

    def __call__(self, data):
        if 'label' in data:
            aux_label = (data['label'].sum(axis=-1, keepdims=True) == 0).astype('uint8')
            data['label'] = np.concatenate([aux_label, data['label']], axis=-1)

        return data
